/*
 * @(#)AddinsConfig.java
 */

import java.util.HashMap;

/**
 * Configure an Addins service of OpenOffice.org Calc.
 * @since 0.1
 * @version 2.0, Dec. 17, 2008
 * @author Naoyuki Tamura
 * @author Hideaki Okamoto
 */
public class AddinsConfig {
    private static final String __serviceName =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins";
    private static HashMap<String,String[][]> functionTable;
    
    static {
	functionTable = new HashMap<String,String[][]>();
	functionTable.
	    put("CVARIABLE", new String[][] {
		    {"CVARIABLE", "Defines a constraint variable."},
		    {"(internal)", "(internal)"},
		    {"infimum", "the lower limit of the variable domain."},
		    {"supremum", "the upper limit of the variable domain."},
		    {"value", "a value of the cell."},
		});
	functionTable.
	    put("CVARIABLES", new String[][] {
		    {"CVARIABLES", "Specifies areas of constraint variables."},
		    {"(internal)", "(internal)"},
		    {"infimum", "the lower limit of the variable domain."},
		    {"supremum", "the upper limit of the variable domain."},
		    {"variables", "a cell range defined as constraint variables."},
		});
	functionTable.
	    put("COBJECTIVE", new String[][] {
		    {"COBJECTIVE", "Specifies the objective variable."},
		    {"(internal)", "(internal)"},
		    {"variable", "a cell defined as an objective variable."},
		});
	functionTable.
	    put("CONSTRAINTS", new String[][] {
		    {"CONSTRAINTS", "Specifies areas of constraints."},
		    {"(internal)", "(internal)"},
		    {"constraints", "a cell range defined as constraints."},
		});
	functionTable.
	    put("CEQUALS", new String[][] {
		    {"CEQUALS", "Declares EQUALS (all equals) constraint."},
		    {"(internal)", "(internal)"},
		    {"variables", "a cell range defined as equals variables."},
		});
	functionTable.
	     put("CNOTEQUALS", new String[][] {
		    {"CNOTEQUALS", "Declares NOTEQUALS (all different) constraint."},
		    {"(internal)", "(internal)"},
		    {"variables", "a cell range defined as all different variables."},
		});
	functionTable.
	     put("CSEQUENTIAL", new String[][] {
		    {"CSEQUENTIAL", "Declares SEQUENTIAL constraint."},
		    {"(internal)", "(internal)"},
		    {"arguments", "arguments."},
		});
	functionTable.
	     put("CSERIALIZED", new String[][] {
		    {"CSERIALIZED", "Defines a SERIALIZED constraint."},
		    {"(internal)", "(internal)"},
		    {"arguments", "arguments."},
		});
	functionTable.
	     put("CTIMEOUT", new String[][] {
		    {"CTIMEOUT", "Sets the timeout."},
		    {"(internal)", "(internal)"},
		    {"timeout", "sets the timeout."},
		});
	functionTable.
	     put("COPTIONS", new String[][] {
		    {"COPTIONS", "Sets the solver options."},
		    {"(internal)", "(internal)"},
		    {"value", "the value of solver option."},
		});
	functionTable.
	     put("CScanDocument", new String[][] {
		    {"CScanDocument", "RESTRICTED!"},
		    {"", ""},
		});
	functionTable.
	     put("CStart", new String[][] {
		    {"CStart", "RESTRICTED!"},
		    {"", ""},
		});
	functionTable.
	     put("CNext", new String[][] {
		    {"CNext", "RESTRICTED!"},
		    {"", ""},
		});
	functionTable.
	     put("CFindBest", new String[][] {
		    {"CFindBest", "RESTRICTED!"},
		    {"", ""},
		});
	functionTable.
	     put("CFunctionName", new String[][] {
		    {"CFunctionName", "RESTRICTED!"},
		    {"", ""},
		});
	functionTable.
	     put("CFunctionArgc", new String[][] {
		    {"CFunctionArgc", "RESTRICTED!"},
		    {"", ""},
		});
	functionTable.
	     put("CFunctionArg", new String[][] {
		    {"CFunctionArg", "RESTRICTED!"},
		    {"", ""},
		});
	functionTable.
	     put("CFunctionArgs", new String[][] {
		    {"CFunctionArgs", "RESTRICTED!"},
		    {"", ""},
		});
	functionTable.
	     put("CAddVariable", new String[][] {
		    {"CAddVariable", "RESTRICTED!"},
		    {"", ""},
		});
	functionTable.
	     put("CAddConstraint", new String[][] {
		    {"CAddConstraint", "RESTRICTED!"},
		    {"", ""},
		});
	functionTable.
	     put("CAreaSize", new String[][] {
		    {"CAreaSize", "RESTRICTED!"},
		    {"", ""},
		});
	functionTable.
	     put("CArea", new String[][] {
		    {"CArea", "RESTRICTED!"},
		    {"", ""},
		});
    }

    protected AddinsConfig() {
    }

    /**
     * Obtains a new instance of an AddinsConfig.
     * @return a new instance of an AddinsConfig.
     */
    public static AddinsConfig newInstance() {
	return new AddinsConfig();
    }
    
    /**
     * Returns the service name.
     * @return the service name.
     */
    public static String serviceName() {
	return __serviceName;
    }

    /**
     * Returns the user-visible function name for an internal name.
     * @param functionName is the name of a method within its interface. 
     * @return the user-visible name of the specified function.
    */
    public static String functionName(String functionName) {
	return functionName;
    }

    /**
     * Returns a description of an addins function.
     * @param functionName is the name of a method within its interface.
     * @return the description of the specified function.
    */
    public static String functionDescription(String functionName) {
	String[][] desc = functionTable.get(functionName);
	return desc == null ? "" : desc[0][1];
    }
  
    /**
     * Return the user-visible name of the <tt>index</tt>-th argument.
     * @param functionName is the name of a method within its interface.
     * @param index the index value of the argument to be returned(0-based).
     * @return the user-visible name of the <tt>index<tt>-th argument.
    */
    public static String argumentName(String functionName, int index) {
	String[][] desc = functionTable.get(functionName);
	if (desc == null || index + 1 < 0 || index + 1 >= desc.length) {
	    return "";
	}
	return desc[index+1][0];
    }

    /**
     * Returns a description of the <tt>index</tt>-th argument.
     * @param functionName is the name of a method within its interface.
     * @param index the index value of the argument to be returned(0-based).
     * @return the description of the <tt>index<tt>-th argument.
     */
    public static String argumentDescription(String functionName, int index) {
	String[][] desc = functionTable.get(functionName);
	if (desc == null || index + 1 < 0 || index + 1 >= desc.length) {
	    return "";
	}
	return desc[index+1][1];
    }
}
