/*
 * @(#)CreamAddins.java
 */
import com.sun.star.lang.Locale;
import com.sun.star.lang.XMultiServiceFactory;
import com.sun.star.lang.XSingleServiceFactory;
import com.sun.star.lang.XServiceName;
import com.sun.star.lang.XServiceInfo;
import com.sun.star.comp.loader.FactoryHelper;
import com.sun.star.registry.XRegistryKey;
import com.sun.star.lib.uno.helper.WeakBase;
import com.sun.star.beans.XPropertySet;
import com.sun.star.sheet.XAddIn;
import com.sun.star.table.XCellRange;
import jp.ac.kobe_u.cs.cream.uno.XCreamAddins;

/**
 * @see CreamSolver
 * @since 0.1
 * @version 0.9, Mar 6 2005
 * @author Naoyuki Tamura
 * @author Hideaki Okamoto
 */
public class CreamAddins {
 
    public static class _CreamAddins extends WeakBase
	implements XCreamAddins, XAddIn, XServiceName, XServiceInfo
    {
        private static final String ADDIN = "com.sun.star.sheet.AddIn";       
        private Locale aFuncLoc_;

	private static final AddinsConfig addinsConfig_ 
	    = AddinsConfig.newInstance();
	private static final String SERVICE = addinsConfig_.serviceName();

	/** TO DO:
	 * This is where you implement all methods of your interface. 
	 * The parameters have to be the same as in your IDL file 
	 * and their types have to be the correct IDL-to-Java mappings 
	 * of their types in the IDL file.
	 * @see XCreamAddins
	 */

	/**
	 * Defines a constraint variable.
	 * @param xOptions 
	 * @param inf is the lower limit of the variable domain.
	 * @param sup is the upper limit of the variable domain.
	 * @param value is the value of the cell.
	 * @return value.
	 */
        public int CVARIABLE(XPropertySet xOptions, 
			     Object inf, Object sup, int value) {
	    return CreamSolver.CVARIABLE(xOptions, inf, sup, value);
	}

	/**
	 * Defines a constraint variables.
	 * @param xOptions 
	 * @param inf is the lower limit of the variable domain.
	 * @param sup is the upper limit of the variable domain.
	 * @param variables is the scope of variables.
	 * @return message string.
	 */
        public Object CVARIABLES(XPropertySet xOptions, 
				 Object inf, Object sup, Object[] variables) {
	    return CreamSolver.CVARIABLES(xOptions, inf, sup, variables);
        }

	/**
	 * Sets the objective variables.
	 * @param xOptions 
	 * @param objective is the objective variable.
	 * @return message string.
	 */
        public Object COBJECTIVE(XPropertySet xOptions, Object objective) {
	    return CreamSolver.COBJECTIVE(xOptions, objective);
        }

	/**
	 * Defines constraints.
	 * @param xOptions 
	 * @param constraints is the scope of constraints.
	 * @return message string.
	 */
        public Object CONSTRAINTS(XPropertySet xOptions, Object[] constraints) {
	    return CreamSolver.CONSTRAINTS(xOptions, constraints);
        }

	/**
	 * Defines a constraint means that variables are all equal.
	 * @param xOptions 
	 * @param variables is the scope of all equal variables.
	 * @return -1 if the variables are all equal, 0 otherwise.
	 */
        public int CEQUALS(XPropertySet xOptions, Object[] variables) {
	    return CreamSolver.CEQUALS(xOptions, variables);
        }
	
	/**
	 * Defines a constraint means that variables are all different.
	 * @param xOptions 
	 * @param variables is the scope of all different variables.
	 * @return -1 if the variables are all different, 0 otherwise.
	 */        
        public int CNOTEQUALS(XPropertySet xOptions, Object[] variables) {
	    return CreamSolver.CNOTEQUALS(xOptions, variables);
        }
         
	/**
	 * Defines a constraint.
	 * @param xOptions 
	 * @param args is the arguments.
	 * @return -1 if the variables are SEQUENTIAL, 0 otherwise.
	 */
        public int CSEQUENTIAL(XPropertySet xOptions, Object[] args) {
	    return CreamSolver.CSEQUENTIAL(xOptions, args);
        }
        
	/**
	 * Defines a constraint.
	 * @param xOptions 
	 * @param args is the arguments.
	 * @return -1 if the variables are SERIALIZED, 0 otherwise.
	 */
        public int CSERIALIZED(XPropertySet xOptions, Object[] args) {
	    return CreamSolver.CSERIALIZED(xOptions, args);
        }        

	/**
	 * Sets the timeout.
	 * @param xOptions 
	 * @param timeout is the timeout.
	 * @return message string.
	 */
	public Object CTIMEOUT(XPropertySet xOptions, double timeout) {
	    return CreamSolver.CTIMEOUT(xOptions, timeout);
	}
                        
	/**
	 * Sets the solver option
	 * @param xOptions 
	 * @param values is the value of a solver option.
	 * @return message string.
	 */
        public Object COPTIONS(XPropertySet xOptions, Object[] values) {
	    return CreamSolver.COPTIONS(xOptions, values);
        }

	/**
	 * Scan the spreadsheet document.
	 * @param document is the spreadsheet document to be scanned.
	 * @param indexArray is the array of the sheet-index to be scanned.
	 */
	public void CScanDocument(Object document, int[] indexArray) {
	    CreamSolver.CScanDocument(document, indexArray);
	}

	/**
	 * Starts the solver, and reflects the solution on the document.
	 * @return true if the solution is available, false otherwise.
	 */
        public boolean CStart() {
	    return CreamSolver.CStart();
        }

	/**
	 * Gets the next solution, and reflects the solution on the document.
	 * @return true if the next solution is available, false otherwise.
	 */
        public boolean CNext() {
	    return CreamSolver.CNext();
        }

	/**
	 * Finds the best solution, and reflects the solution on the document.
	 * @return true if the solution is available, false otherwise.
	 */
	public boolean CFindBest() {
	    return CreamSolver.CFindBest();
	}

	/**
	 * An additional method for StarSuite BASIC programming.
	 * @param function is a function string.
	 * @return a function name string. 
	 */
	public String CFunctionName(String function) {
	    return CreamSolver.CFunctionName(function);
	}

	/**
	 * An additional method for StarSuite BASIC programming.
	 * @param function is a function string.
	 * @return the number of the function arguments.
	 */
	public int CFunctionArgc(String function) {
	    return CreamSolver.CFunctionArgc(function);
	}

	/**
	 * An additional method for StarSuite BASIC programming.
	 * @param function is a function string.
	 * @param index is the index of the argument to be returned.
	 * @return the string value of <tt>index</tt>-th argument.
	 */
	public String CFunctionArg(String function, int index) {
	    return CreamSolver.CFunctionArg(function, index);
	}
	
	/**
	 * An additional method for StarSuite BASIC programming.
	 * @param function is a function string.
	 * @return string array of the function arguments .
	 */
	public String[] CFunctionArgs(String function) {
	    return CreamSolver.CFunctionArgs(function);
	}

	/** 
	 * An additional method for StarSuite BASIC programming.
	 * @param inf is the lower limit of the variable domain.
	 * @param sup is the upper limit of the variable domain.
	 * @return a key to identify the variable.
	 */
	public String CAddVariable(Object inf, Object sup) {
	    return CreamSolver.CAddVariable(inf, sup);
	}

	/** 
	 * An additional method for StarSuite BASIC programming.
	 * @param constraint is the constraint string.
	 */
	public void CAddConstraint(String constraint) {
	    CreamSolver.CAddConstraint(constraint);
	}

	/**
	 * An additional method for StarSuite BASIC programming.
	 * @return the number of the ranges defined by CONSTRAINTS.
	 */
	public int CAreaSize() {
	    return CreamSolver.CAreaSize();
	}

	/**
	 * An additional method for StarSuite BASIC programming.
	 * @param index is the index of the range.
	 * @return <tt>index</tt>-th range name defined by CONSTRAINTS.
	 */
	public String CArea(int index) {
	    return CreamSolver.CArea(index);
	}

        /**
	 * Returns the service name used to create this object by a factory.
	 * @return the service name used to create this object by a factory.
	 * @see com.sun.star.io.XPersistObject
	 */
        public String getServiceName() { return( SERVICE ); }
        
        /** 
	 * Tests whether the specified service is supported.
	 * @param stringServiceName name of service to be tested.
	 * @return true, if service is supported, false otherwise.
	 * @see com.sun.star.lang.XServiceInfo
	 */
	public boolean supportsService(String stringServiceName) {
            return( stringServiceName.equals( ADDIN ) ||
                    stringServiceName.equals( SERVICE ) );
        }
        
	/**
	 * Returns the implementation name of the service implementation.
	 * @return unique name of the implementation.
	 * @see com.sun.star.lang.XServiceInfo
	 */
        public String getImplementationName() {
            return( _CreamAddins.class.getName() );
        }
        
	/**
	 * Returns the supported service names of the implementation.
	 * @return sequence of service names that are supported.
	 * @see com.sun.star.lang.XServiceInfo
	 */
        public String[] getSupportedServiceNames() {
            String[] stringSupportedServiceNames = { ADDIN, SERVICE };
            return( stringSupportedServiceNames );
        }

	/**
	 * Returns the internal name for an user-visible name.
	 * @param displayName is the user-visible name of a function.
	 * @return the exact name of the method within XCreamAddins interface.
	 * @see com.sun.star.sheet.XAddIn
	 */                
        public String getProgrammaticFuntionName(String displayName) {
            return( "" );
        }
        
	/**
	 * Returns the user-visible function name for an internal name.
	 * @param functionName is the name of a method within XCreamAddins.
	 * @return the user-visible name of the specified function.
	 * @see com.sun.star.sheet.XAddIn
	 */
	public String getDisplayFunctionName(String functionName) {
	    return addinsConfig_.functionName(functionName);
	}

	/**
	 * Returns the description of a function.
	 * @param functionName is the name of a method within XCreamAddins.
	 * @return the description of the specified function.
	 * @see com.sun.star.sheet.XAddIn
	 */
	public String getFunctionDescription(String functionName) {
	    return addinsConfig_.functionDescription(functionName);
	}

	/**
	 * Returns the user-visible name of the <tt>index</tt>-th argument.
	 * @param functionName is the name of a method within XCreamAddins.
	 * @param index is the index of the argument(0-based).
	 * @return the user-visible name of the <tt>index</tt>-th argument.
	 * @see com.sun.star.sheet.XAddIn
	*/
	public String getDisplayArgumentName(String functionName, int index) {
	   return addinsConfig_.argumentName(functionName, index);
	}
	
	/**
	 * Returns the description of the <tt>index</tt>-th argument.
	 * @param functionName is the name of a method within XCreamAddins.
	 * @param index is the index of the argument(0-based).
	 * @return the description of the <tt>index</tt>-th argument.
	 * @see com.sun.star.sheet.XAddIn
	 */
	public String getArgumentDescription(String functionName, int index) {
	    return addinsConfig_.argumentDescription(functionName, index);
	}

	/**
	 * Returns the name of the category the function belongs to.
	 * @param functionName is the name of a method within XCreamAddins.
	 * @return the category name the specidied function belongs to.
	 * @see com.sun.star.sheet.XAddIn
	 */	        
        public String getProgrammaticCategoryName(String functionName) {
            return( "Add-In" );
        }
        
	/**
	 * Returns the visible name of the category the function belongs to.
	 * @param functionName is the name of a method within XCreamAddins.
	 * @return the visible category name the specified function belongs to.
	 * @see com.sun.star.sheet.XAddIn 
	*/
        public String getDisplayCategoryName(String functionName) {
            return( "Add-In" );
        }

	/**
	 * Sets the locale to be used by this object.
	 * @param locale is the locale to be used by this object.
	 * @see com.sun.star.lang.XLocalizable
	*/
        public void setLocale(Locale locale) { aFuncLoc_ = locale; }
                
	/**
	 * Gets the locale which is used by this object.
	 * @return the locale, which is used by this object.
	 * @see com.sun.star.lang.XLocalizable
	 */
        public Locale getLocale() { return( aFuncLoc_ ); }
    }
    
    /**
     * Returns a factory for creating the service.
     * @return a factory for creating the component.
     * @param implName is the name of the implementation. 
     * @param multiFactory is the service manager to be used if needed.
     * @param regKey is the registryKey.
     * @see com.sun.star.comp.loader.JavaLoader
     */
    public static XSingleServiceFactory __getServiceFactory
	(String implName, XMultiServiceFactory mFactory, XRegistryKey regKey) 
    {
        XSingleServiceFactory factory = null;
        if (implName.equals(_CreamAddins.class.getName()))
            factory = FactoryHelper.getServiceFactory
		(_CreamAddins.class, _CreamAddins.SERVICE, mFactory, regKey);
        return factory;
    }
    
    /**
     * Writes the service information into the given registry key.
     * @return true if the operation succeeded, false otherwise.
     * @param regKey is the registryKey.
     * @see com.sun.star.comp.loader.JavaLoader
     */
    public static boolean __writeRegistryServiceInfo(XRegistryKey regKey) 
    {
        return FactoryHelper.writeRegistryServiceInfo
	    (_CreamAddins.class.getName(), _CreamAddins.SERVICE, regKey)
	    && FactoryHelper.writeRegistryServiceInfo
	    (_CreamAddins.class.getName(), _CreamAddins.ADDIN, regKey);
    }   
}
