/*
 * @(#)CreamSolver.java
 */

// import com.sun.star.sheet.*;
import com.sun.star.sheet.XSpreadsheetDocument;
import com.sun.star.sheet.XSpreadsheets;
import com.sun.star.sheet.XSpreadsheet;
import com.sun.star.sheet.XCellAddressable;
import com.sun.star.sheet.XCellRangeAddressable;
import com.sun.star.sheet.XCellRangesQuery;
import com.sun.star.sheet.CellFlags;

// import com.sun.star.table.*; 
import com.sun.star.table.XCellRange;
import com.sun.star.table.XCell;
import com.sun.star.table.CellContentType;
import com.sun.star.table.CellAddress;
import com.sun.star.table.CellRangeAddress;

import com.sun.star.beans.XPropertySet;
import com.sun.star.beans.XPropertyChangeListener;
import com.sun.star.uno.UnoRuntime;
import com.sun.star.uno.Any;
import com.sun.star.uno.AnyConverter;
import com.sun.star.uno.Type;
import com.sun.star.container.XNamed;
import com.sun.star.container.XNameAccess;
import com.sun.star.container.XIndexAccess;
import com.sun.star.container.XEnumerationAccess;
import com.sun.star.container.XEnumeration;

// import java.util.*;
import java.util.List;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

//import java.util.regex.*;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import jp.ac.kobe_u.cs.cream.*;

/**
 *
 * @since 0.1
 * @version 0.9, Mar 6 2005
 * @author Naoyuki Tamura
 * @author Hideaki Okamoto
 */
public class CreamSolver {
    private static _CreamSolver solver_ = null;

    private static String _unoToString(Object uno) {
	String s = "";
	try {
	    if (AnyConverter.isString(uno)) s = AnyConverter.toString(uno);
	    else s = Integer.toString(_unoToInt(uno)); 
	} catch (com.sun.star.lang.IllegalArgumentException e) {
	    e.printStackTrace();
	}
	return s;
    }

    private static int _unoToInt(Object uno) {
	double x = 0.0;
	try {
	    if (AnyConverter.isInt(uno)) x = AnyConverter.toInt(uno);
	    else if (AnyConverter.isShort(uno)) x = AnyConverter.toShort(uno);
	    else if (AnyConverter.isLong(uno)) x = AnyConverter.toLong(uno);
	    else if (AnyConverter.isFloat(uno)) x = AnyConverter.toFloat(uno);
	    else x = AnyConverter.toDouble(uno);
	} catch (com.sun.star.lang.IllegalArgumentException e) {
	    e.printStackTrace();
	}
	return (int)x;
    }

    private static void _addUnoToList(List<Object> list, Object uno) {
	try {
	    if (AnyConverter.isArray(uno)) {
		Object[] unoArray = (Object[])(AnyConverter.toArray(uno));
		
		for (int i = 0; i < unoArray.length; i++) 
		    _addUnoToList(list, unoArray[i]);
	    } else { list.add(uno); }
	} catch (com.sun.star.lang.IllegalArgumentException e) {
	    e.printStackTrace();
	}
    }

    /**
     * Defines a constraint variable.
     * @param xOptions
     * @param inf is the lower limit of the variable domain.
     * @param sup is the upper limit of the variable domain.
     * @param value is a value of the cell.
     * @return value.
     */
    public static int CVARIABLE(XPropertySet xOptions, 
				Object inf, Object sup, int value) {
	if (_unoToInt(sup) < _unoToInt(inf)) 
	    throw new IllegalArgumentException();
	return value;
    }

    /**
     * Defines a constraint variables.
     * @param xOptions
     * @param inf is the lower limit of the variable domain.
     * @param sup is the upper limit of the variable domain.
     * @param vars is the scope of variables.
     * @return message string.
     */
    public static Object CVARIABLES(XPropertySet xOptions,
				    Object inf, Object sup, Object[] vars) {
	if (_unoToInt(sup) < _unoToInt(inf))
	    throw new IllegalArgumentException();
	return "CVARIABLES[" + _unoToString(inf) + 
	    ":" + _unoToString(sup) + "]";
    }
    
    /**
     * Sets the objective variables.
     * @param xOptions
     * @param objective is the objective variable.
     * @return message string.
     */
    public static Object COBJECTIVE(XPropertySet xOptions, 
				    Object objective) {
	return "OBJECTIVE[" + "]";
    }

    /**
     * Defines constraints.
     * @param xOptions
     * @param constrains is the scope of constraints.
     * @return message string.
     */    
    public static Object CONSTRAINTS(XPropertySet xOptions, 
				     Object[] constraints) {
	return "CONSTRAINTS[" + "]";
    }    

    /**
     * Defines a constraint means that variables are all equal.
     * @param xOptions
     * @param vars is the scope of all equal variables.
     * @return -1 if the vars are all equal, 0 otherwise.
     */    
    public static int CEQUALS(XPropertySet xOptions, Object[] vars) {
	List<Object> list = new ArrayList<Object>();
	for (int i = 0; i < vars.length; i++) _addUnoToList(list, vars[i]);

	return _isEqSatisfied(list.toArray())? -1 : 0;
    }

    private static boolean _isEqSatisfied(Object[] vars) {
	boolean eq = true;	
	for (int i = 0; eq && i < vars.length; i++) {
	    int x = _unoToInt(vars[i]);
	    for (int j = i + 1; eq && j < vars.length; j++) {
		if (x != _unoToInt(vars[j])) eq = false;
	    }
	}
	return eq;
    }
    
    /**
     * Defines a constraint means that variables are all different.
     * @param xOptions
     * @param vars is the scope of all different variables.
     * @return -1 if the vars are all different, 0 otherwise.
     */    
    public static int CNOTEQUALS(XPropertySet xOptions, Object[] vars) {
	List<Object> list = new ArrayList<Object>();
	for (int i = 0; i < vars.length; i++) _addUnoToList(list, vars[i]);

	return _isNeqSatisfied(list.toArray())? -1 : 0;
    }
    
    private static boolean _isNeqSatisfied(Object[] vars) {
	boolean neq = true;
	for (int i = 0; neq && i < vars.length; i++) {
	    int x = _unoToInt(vars[i]);
	    for (int j = i + 1; neq && j < vars.length; j++) {
		if (x == _unoToInt(vars[j])) neq = false;
	    }
	}
	return neq;
    }

     /**
     * Defines a constraint.
     * @param xOptions
     * @param args is the arguments.
     * @return -1 if the vars are SEQUENTIAL, 0 otherwise.
     */
    public static int CSEQUENTIAL(XPropertySet xOptions, Object[] args) {
	List<Object> list = new ArrayList<Object>();
	for (int i = 0; i < args.length; i++) _addUnoToList(list, args[i]);
	if (list.size() % 2 == 0)
	    throw new IllegalArgumentException();
	
	return _isSeqSatisfied(list.toArray())? -1 : 0;
    }
    
    private static boolean _isSeqSatisfied(Object[] args) {
	boolean seq = true;
	for (int i = 0; seq && i < args.length / 2; i++) {
	    int x = _unoToInt(args[i + 1]) - _unoToInt(args[i]);
	    int y = _unoToInt(args[i + args.length / 2 + 1]);
	    if (x < y) seq = false; 
	}
	return seq;
    }
    
    /**
     * Defines a constraint.
     * @param xOptions
     * @param args is the arguments.
     * @return -1 if the vars are SERIALIZED, 0 otherwise.
     */
    public static int CSERIALIZED(XPropertySet xOptions, Object[] args) {
	List<Object> list = new ArrayList<Object>();
	for (int i = 0; i < args.length; i++) _addUnoToList(list, args[i]);
	if (list.size() % 2 != 0)
	    throw new IllegalArgumentException();

	return _isSerSatisfied(list.toArray())? -1 : 0;
    }

    private static boolean _isSerSatisfied(Object[] args) {
	boolean ser = true;
	for (int i = 0; ser && i < args.length / 2; i++) {
	    int domainInf1 = _unoToInt(args[i]);
	    int domainSup1 = _unoToInt(args[i]) 
		+ _unoToInt(args[i + args.length / 2]);
	    for (int j = i + 1; ser && j < args.length / 2; j++) {
		int domainInf2 = _unoToInt(args[j]);
		int domainSup2 = _unoToInt(args[j]) 
		    + _unoToInt(args[j + args.length / 2]);
		
		if (!(domainSup1 <= domainInf2 || domainSup2 <= domainInf1))
		    ser = false;
	    }
	}
	return ser;
    }
    
    /**
     * Sets the timeout.
     * @param xOptions
     * @param timeout is the timeout.
     * @return message string.
     */
    public static Object CTIMEOUT(XPropertySet xOptions, double timeout) {
	return "TIMEOUT[" + Double.toString(timeout) + "(sec)]";
    }

    /**
     * Sets the solver option.
     * @param xOptions
     * @param values is the value of a solver option. 
     * @return message string.
     */
    public static Object COPTIONS(XPropertySet xOptions, Object[] values) {
	String s = "OPTIONS[";
	String delim = "";

	for (int i = 0; i < values.length; i++) { 
	    s += delim + _unoToString(values[i]).toUpperCase();
	    delim = ",";
	}
	return s + "]";
    }
     
    /**
     * Scans the spreadsheet document, and add constraints to the solver.
     * @param document is the spreadsheet document to be scanned.
     * @param indexArray is the array of the sheet-index to be scanned.
     */
    public static void CScanDocument(Object document, int[] indexArray) {
	if (solver_ != null) solver_.stop();

	XSpreadsheetDocument doc = (XSpreadsheetDocument)UnoRuntime.
	    queryInterface(XSpreadsheetDocument.class, document);
	
	solver_ = new _CreamSolver(doc, indexArray);
	solver_.scanDocument();
    }

    /**
     * Starts the solver, and reflects the solution on the document.
     * @return true if the solution is available, false otherwise.
     */
    public static boolean CStart() {
	if (solver_ == null) return false;

	solver_.start();
	solver_.waitNext();

	Solution solution = solver_.getSolution();
	if (solution == null) {
	    solver_.stop();
	    solver_ = null;
	    return false;
	}
	solver_.setSolution(solution);
	return true;
    }
    
    /**
     * Gets the next solution, and reflects the solution on the document.
     * @return true if the next solution is available, false otherwise.
     */
    public static boolean CNext() {
	if (solver_ == null) return false;
	
	solver_.resume();
	solver_.waitNext();

	Solution solution = solver_.getSolution();
	if (solution == null) {
	    solver_.stop();
	    solver_ = null;
	    return false;
	}
	solver_.setSolution(solution);
	return true;
    }

    /**
     * Finds the best solution, and reflects the solution on the document.
     * @return true if the solution is available, false otherwise.
     */
    public static boolean CFindBest() {
	if (solver_ == null) return false;
	
	Solution solution = solver_.findBest();
	if (solution == null) {
	    solver_.stop();
	    solver_ = null;
	    return false;
	}
	solver_.setSolution(solution);
	return true;
    }

    public static String CFunctionName(String function) {
	return solver_.CFunctionName(function);
    }
    
    public static int CFunctionArgc(String function) {
	return solver_.CFunctionArgc(function);
    }

    public static String CFunctionArg(String function, int index) {
	return solver_.CFunctionArg(function, index);
    }

    public static String[] CFunctionArgs(String function) {
	return solver_.CFunctionArgs(function);
    }

    public static String CAddVariable(Object inf, Object sup) {
	if (_unoToInt(sup) < _unoToInt(inf)) 
	    throw new IllegalArgumentException();
	return solver_.CAddVariable(_unoToString(inf), _unoToString(sup));
    }

    public static void CAddConstraint(String constraint) {
	solver_.CAddConstraint(constraint);
    }

    public static int CAreaSize() { return solver_.CAreaSize(); }

    public static String CArea(int index) { return solver_.CArea(index); }
} /* End of class CreamSolver*/

class _CreamSolver {
    private static final String CREAM_CELL_REGEX = "^\\s*=(\\s*jp\\.ac\\.kobe_u\\.cs\\.cream\\.uno\\.CreamAddins\\.\\w*\\s*\\(.*\\)\\s*)$";
    private static final String VARIABLE =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.CVARIABLE".toUpperCase();
    private static final String VARIABLES =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.CVARIABLES".toUpperCase();
    private static final String OBJECTIVE =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.COBJECTIVE".toUpperCase();    
    private static final String CONSTRAINTS =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.CONSTRAINTS".toUpperCase();
    private static final String EQUALS =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.CEQUALS".toUpperCase();
    private static final String NOTEQUALS =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.CNOTEQUALS".toUpperCase();
    private static final String SEQUENTIAL =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.CSEQUENTIAL".toUpperCase();
    private static final String SERIALIZED =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.CSERIALIZED".toUpperCase();
    private static final String TIMEOUT =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.CTIMEOUT".toUpperCase();
    private static final String OPTIONS =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.COPTIONS".toUpperCase();

    private XSpreadsheetDocument doc_ = null;
    private int[] indexArray_ = null;
    private Hashtable<String, CellVariable> variables_;

    private Network network_;
    private Solver solver_ = null;
    private Variable objective_ = null;
    private int options_ = Solver.NONE;
    private long timeout_ = -1;
    private String[] solvers_ = null;

    private RuntimeLogger logger_ = null;
    private static final boolean LOGGER_ACTIVE = false;

    private InternalVariableKeeper internal_; // For BASIC
    private List<String> constraintAreaList_ = null; // For BASIC
        
    class CellVariable {
	private XCell cell_;
	private Variable variable_;

	public CellVariable(XCell cell, Variable v) {
	    this.cell_ = cell;
	    this.variable_ = v;
	}

	public void setCell(XCell cell) { this.cell_ = cell; }
	public XCell getCell() { return this.cell_; }
	
	public void setVariable(Variable v) { this.variable_ = v; }
	public Variable getVariable() { return this.variable_; }
    }

    /**
     * constructor.
     * @param doc is the document to be scanned.
     * @param indexArray is the array of the sheet-index to be scanned.
     */
    public _CreamSolver(XSpreadsheetDocument doc, int[] indexArray) {
	logger_ = new RuntimeLogger(LOGGER_ACTIVE);

	this.doc_ = doc;
	this.indexArray_ = indexArray;

	this.network_ = new Network();
	this.variables_ = new Hashtable<String, CellVariable>();

	this.internal_ = new InternalVariableKeeper(); // For BASIC
	this.constraintAreaList_ = new ArrayList<String>(); // For BASIC
    }

    private XSpreadsheet _getSheet(int index) {
	XSpreadsheets sheets = doc_.getSheets();
	XSpreadsheet sheet = null;
	XIndexAccess access = (XIndexAccess)UnoRuntime.
	    queryInterface(XIndexAccess.class, sheets);
	try {
	    sheet = (XSpreadsheet)UnoRuntime.
		queryInterface(XSpreadsheet.class, access.getByIndex(index));
	} catch (com.sun.star.lang.WrappedTargetException e) {
	    e.printStackTrace();
	} catch (com.sun.star.lang.IndexOutOfBoundsException e) {
	    e.printStackTrace();
	}
	return sheet;
    }

    private XSpreadsheet _getSheet(String name) {
	XSpreadsheets sheets = doc_.getSheets();
	XSpreadsheet sheet = null;
	XNameAccess access = (XNameAccess)UnoRuntime.
	    queryInterface(XNameAccess.class, sheets);
	try {
	    sheet = (XSpreadsheet)UnoRuntime.
		queryInterface(XSpreadsheet.class, access.getByName(name));
	} catch (com.sun.star.lang.WrappedTargetException e) {
	    e.printStackTrace();
	} catch (com.sun.star.container.NoSuchElementException e) {
	    e.printStackTrace();
	}
	return sheet;
    }

    private XCellRange _getWholeRange(XSpreadsheet sheet) {
	XCellRange range = (XCellRange)UnoRuntime.
	    queryInterface(XCellRange.class, sheet);
	return range;
    }
    
    private XCellRange _getWholeRange(XCellRange range) {
	XCellRangeAddressable address = (XCellRangeAddressable)UnoRuntime.
	    queryInterface(XCellRangeAddressable.class, range);
	int sheet_index = address.getRangeAddress().Sheet;
	return _getWholeRange(_getSheet(sheet_index));
    }

    private XCellRange _getWholeRange(XCell cell) {
	XCellAddressable address = (XCellAddressable)UnoRuntime.
	    queryInterface(XCellAddressable.class, cell);
	int sheet_index = address.getCellAddress().Sheet;
	return _getWholeRange(_getSheet(sheet_index));
    }
    
    private List<XCell> _creamCellList(XCellRange range) {
	List<XCell> list = new ArrayList<XCell>();
	XEnumeration xenum = _cellRangeEnumeration(range);
	Pattern pattern = Pattern.compile(CREAM_CELL_REGEX);
	try {
	    while (xenum.hasMoreElements()) {
		XCell cell = (XCell)UnoRuntime.
		    queryInterface(XCell.class, xenum.nextElement());
		String formula = cell.getFormula();
		Matcher matcher = pattern.matcher(formula);
		if (matcher.matches()) list.add(cell);
	    }
	} catch (com.sun.star.container.NoSuchElementException e) {
	    e.printStackTrace();
	} catch (com.sun.star.lang.WrappedTargetException e) {
	    e.printStackTrace();
	}
	return list;
    }

    private void _scanVariable(XCellRange range) {
	List<XCell> cells = _creamCellList(range);
	for (Iterator ite = cells.iterator(); ite.hasNext(); ) {
	    XCell cell = (XCell)ite.next();
	    CalcParser.Expression x = 
		new CalcParser(cell.getFormula()).parse();
	    try {
		if (x.isFunctor(VARIABLE)) {
		    int inf = _domainValue(range, x.args[0]);
		    int sup = _domainValue(range, x.args[1]);
		    String key = _getCellKey(cell);
		    
		    if (variables_.get(key) == null && inf <= sup) {
			Variable v = new IntVariable(network_, inf, sup);
			variables_.put(key, new CellVariable(cell, v));
			logger_.log("variable " + key + " : "
				    + String.valueOf(inf) + ","
				    + String.valueOf(sup));
		    }
		}
	    } catch (java.lang.IllegalArgumentException e) {
		e.printStackTrace();
	    }
	}
    }

    private int _domainValue(XCellRange current, CalcParser.Expression x) {
	int value = 0;
	String s = x.op.str.toUpperCase();
	if (s.equals("MIN") || s.equals("MINIMUM")) {
	    value = IntDomain.MIN_VALUE;
	} else if (s.equals("MAX") || s.equals("MAXIMUM")) {
	    value = IntDomain.MAX_VALUE;
	} else {
	    value = _cellValue(current, x);
	}
	return value;
    }

    private XCell _getCell(XCellRange current, CalcParser.Expression x) {
	XCell cell = null;
	try {
	    if (x.isSymbol())
		cell = _getTargetRange(current, x).getCellByPosition(0, 0);
	} catch (com.sun.star.lang.IndexOutOfBoundsException e) {
	    e.printStackTrace();
	}
	return cell;
    }

    private XCellRange _getTargetRange(XCellRange current, 
				       CalcParser.Expression x) {
	XCellRange range = current;
	Pattern pattern = Pattern.compile(".*\\..*");
	Matcher matcher = pattern.matcher(x.op.str);
	if (matcher.matches()) {
	    Pattern p = Pattern.compile("\\.");
	    String[] ss = p.split(x.op.str);
	    range = _getWholeRange(_getSheet(ss[0]));
	}
	return range.getCellRangeByName(x.op.str);
    }

    private int _cellValue(XCellRange current, CalcParser.Expression x) {
	int value = 0;
	if (x.isInteger()) {
	    value = Integer.parseInt(x.op.str);
	} else if (x.isSymbol()) {
	    XCell xCell = _getCell(current, x);
	    value = (int)xCell.getValue();
	} else if (x.isOperator("+", 1)) {
	    value = _cellValue(current, x.args[0]);
	} else if (x.isOperator("+", 2)) { 
	    value = _cellValue(current, x.args[0]) + 
		_cellValue(current, x.args[1]);
	} else if (x.isOperator("-", 1)) {
	    value = -1 * _cellValue(current, x.args[0]);
	} else if (x.isOperator("-", 2)) { 
	    value = _cellValue(current, x.args[0]) -
		_cellValue(current, x.args[1]);
	} else if (x.isOperator("*", 2)) {
	    value = _cellValue(current, x.args[0]) * 
		_cellValue(current, x.args[1]);
	} else if (x.isFunctor("ABS")) {
	    value = Math.abs(_cellValue(current, x.args[0]));
	} else if (x.isFunctor("SUM")) {
	    value = _sumValue(current, x.args);
	}
	return value;
    }
    
    private int _sumValue(XCellRange current, CalcParser.Expression[] args) {
	int sum = 0;
	if (args != null) {
	    for(int i = 0; i < args.length; i++) {
		if (args[i].isSymbol()) {
		    XCellRange range = _getTargetRange(current, args[i]);
		    XEnumeration xenum = _cellRangeEnumeration(range);
		    try {
			while (xenum.hasMoreElements()) {
			    XCell cell = (XCell)UnoRuntime.
				queryInterface(XCell.class, xenum.nextElement());
			    sum += (int)cell.getValue();
			}
		    } catch (com.sun.star.container.NoSuchElementException e) {
			e.printStackTrace();
		    } catch (com.sun.star.lang.WrappedTargetException e) {
			e.printStackTrace();
		    }
		} else {
		    sum += _cellValue(current, args[i]);
		}
	    }
	}	    	 
	return sum;
    }

    private void _look_ahead(XCellRange range) {
	List<XCell> cells = _creamCellList(range);
	for (Iterator ite = cells.iterator(); ite.hasNext(); ) {
	    XCell cell = (XCell)ite.next();
	    CalcParser.Expression x = 
		new CalcParser(cell.getFormula()).parse();
	    try {
		if (x.isFunctor(VARIABLES)) {
		    int inf = _domainValue(range, x.args[0]);
		    int sup = _domainValue(range, x.args[1]);
		    _look_aheadVariables(range, inf, sup, x.args, 2);
		}
	    } catch (java.lang.IllegalArgumentException e) {
		e.printStackTrace();
	    }
	}	
    }

    private void _look_aheadVariables(XCellRange current, int inf, int sup,
				      CalcParser.Expression[] args, int i0) {
	for (int i = i0; i < args.length; i++) {
	    XCellRange range = _getTargetRange(current, args[i]);
	    XEnumeration xenum = _cellRangeEnumeration(range);
	    try {
		while (xenum.hasMoreElements()) {
		    XCell cell = (XCell)UnoRuntime.
			queryInterface(XCell.class, xenum.nextElement());
		    String key = _getCellKey(cell);
		    if (variables_.get(key) == null && inf <= sup) {
			Variable v = new IntVariable(network_, inf, sup);
			variables_.put(key, new CellVariable(cell, v));
			logger_.log("variable " + key + " : "
				    + String.valueOf(inf) + ","
				    + String.valueOf(sup));
		    }
		}
	    } catch (com.sun.star.container.NoSuchElementException e) {
		e.printStackTrace();
	    } catch (com.sun.star.lang.WrappedTargetException e) {
		e.printStackTrace();
	    }
	}
    }

    private void _scanVariables(XCellRange range) {
	List<XCell> cells = _creamCellList(range);
	for (Iterator ite = cells.iterator(); ite.hasNext(); ) {
	    XCell cell = (XCell)ite.next();
	    CalcParser.Expression x = 
		new CalcParser(cell.getFormula()).parse();
	    try {
		if (x.isFunctor(VARIABLES)) {
		    _variablesFunction(range, x.args);
		} else if (x.isFunctor(CONSTRAINTS)) {
		    _constraintVariables(range, x.args);
		}
	    } catch (java.lang.IllegalArgumentException e) {
		e.printStackTrace();
	    }
	}	
    }
    
    private void _variablesFunction(XCellRange current,
				    CalcParser.Expression[] args) {
	for (int i = 2; i < args.length; i++) {
	    XCellRange range = _getTargetRange(current, args[i]);
	    XEnumeration xenum = _cellRangeEnumeration(range);
	    try {
		while (xenum.hasMoreElements()) {
		    XCell cell = (XCell)UnoRuntime.
			queryInterface(XCell.class, xenum.nextElement());
		    String key = _getCellKey(cell);

		    if (variables_.get(key) != null) {
			Variable v1 = 
			    ((CellVariable)variables_.get(key)).getVariable();
			if (cell.getType() == CellContentType.FORMULA) {
			    CalcParser.Expression exp = 
				new CalcParser(cell.getFormula()).parse();
			    Variable v2 = 
				_getTerm(_getWholeRange(range), exp);
			    if (v2 != null) 
				new Equals(network_, v1, v2);
			}
		    }
		}
	    } catch (com.sun.star.container.NoSuchElementException e) {
		e.printStackTrace();
	    } catch (com.sun.star.lang.WrappedTargetException e) {
		e.printStackTrace();
	    }
	}
    }

    private void _constraintVariables(XCellRange current,
				    CalcParser.Expression[] args) {
	for (int i = 0; i < args.length; i++) {
	    XCellRange range = _getTargetRange(current, args[i]);
	    XEnumeration xenum = _cellRangeEnumeration(range);
	    try {
		while (xenum.hasMoreElements()) {
		    XCell cell = (XCell)UnoRuntime.
			queryInterface(XCell.class, xenum.nextElement());
		    String key = _getCellKey(cell);
		    if (variables_.get(key) != null)
			continue;
		    if (cell.getType() == CellContentType.FORMULA) {
			CalcParser.Expression exp = 
			    new CalcParser(cell.getFormula()).parse();
			Variable v = _getTerm(_getWholeRange(range), exp);
			if (v != null) {
			    variables_.put(key, new CellVariable(cell, v));
			}
		    }
		}
	    } catch (com.sun.star.container.NoSuchElementException e) {
		e.printStackTrace();
	    } catch (com.sun.star.lang.WrappedTargetException e) {
		e.printStackTrace();
	    }
	}
    }

    private Variable[] _getTerms(XCellRange current, 
				 CalcParser.Expression[] args) {
	List<Variable> list = new ArrayList<Variable>();
	for (int i = 0; i < args.length; i++) {
	    Variable v = _getTerm(current, args[i]);
	    if (v != null) list.add(v);
	}
	Variable[] vs = new Variable[list.size()];
	vs = (Variable[])list.toArray(vs);
	return vs;
    }

    private Variable _getTerm(XCellRange current, CalcParser.Expression x) {
	Variable v = null;
	if (x.isInteger()) {
	    v = new IntVariable(network_, Integer.parseInt(x.op.str));
	} else if (x.isSymbol()) {
	    if (internal_.isInternalVariableKey(x.toString())) {
		v = internal_.getVariable(x.toString()); // For BASIC
	    } else {
		v = _getCellVariable(_getCell(current, x));
	    }
	} else if (x.isOperator("-", 1)) {
	    Variable[] vs = _getTerms(current, x.args);
	    v = new IntVariable(network_);
	    new IntFunc(network_, IntFunc.NEGATE, v, vs[0]);
	} else if (x.isOperator("+", 2)) {
	    Variable[] vs = _getTerms(current, x.args);
	    v = new IntVariable(network_);
	    new IntArith(network_, IntArith.ADD, v, vs[0], vs[1]);
	} else if (x.isOperator("-", 2)) {
	    Variable[] vs = _getTerms(current, x.args);
	    v = new IntVariable(network_);
	    new IntArith(network_, IntArith.SUBTRACT, v, vs[0], vs[1]);
	} else if (x.isOperator("*", 2)) {
	    Variable[] vs = _getTerms(current, x.args);
	    v = new IntVariable(network_);
	    new IntArith(network_, IntArith.MULTIPLY, v, vs[0], vs[1]);
	} else if (x.isFunctor("ABS")) {
	    Variable[] vs = _getTerms(current, x.args);
	    v = new IntVariable(network_);
	    new IntFunc(network_, IntFunc.ABS, v, vs[0]);
	} else if (x.isFunctor("SUM")) {
	    v = _sum(current, x.args);
	} 
	return v;
    }

    private Variable _sum(XCellRange current, CalcParser.Expression[] args) {
	if (args == null) return new IntVariable(network_, 0);
	
	IntVariable sum = null;
	for (int i = 0; i < args.length; i++) {
	    if(args[i].isSymbol()) {
		XCellRange range = _getTargetRange(current, args[i]);
		XEnumeration xenum = _cellRangeEnumeration(range);
		try {
		    while (xenum.hasMoreElements()) {
			XCell cell = (XCell)UnoRuntime.
			    queryInterface(XCell.class, xenum.nextElement());
			
			IntVariable v = (IntVariable)_getCellVariable(cell);
			
			if (v != null)
			    sum = (sum == null) ? v : sum.add(v);
		    }
		} catch (com.sun.star.container.NoSuchElementException e) {
		    e.printStackTrace();
		} catch (com.sun.star.lang.WrappedTargetException e) {
		    e.printStackTrace();
		}
	    } else {
		int value = _cellValue(current, args[i]);
		IntVariable v = new IntVariable(network_, value);
		sum = (sum == null)? v : sum.add(v);		
	    }
	}
	return sum;
    }

    private void _scanConstraints(XCellRange range) {
	List<XCell> cells = _creamCellList(range);
	for (Iterator ite = cells.iterator(); ite.hasNext(); ) {
	    XCell cell = (XCell)ite.next();
	    CalcParser parser = new CalcParser(cell.getFormula());

	    try {
		CalcParser.Expression x = parser.parse();
	
		if (x.isFunctor(CONSTRAINTS)) {
		    _constraintsFunction(range, x.args);
		} else if (x.isFunctor(OPTIONS)) {
		    _optionsFunction(x.args);
		} else if (x.isFunctor(TIMEOUT, 1)) {
		    _timeoutFunction(x.args);
		} else if (x.isFunctor(OBJECTIVE)) {
		    _objectiveFunction(range, x.args[0]);
		}
	    } catch (java.lang.IllegalArgumentException e) {
		e.printStackTrace();
	    }
	}
    }

    private String _getAbsoluteRangeName(XCellRange current,
					 CalcParser.Expression x) {
	String s = x.op.str;
	Pattern pattern = Pattern.compile(".*\\..*");
	Matcher matcher = pattern.matcher(x.op.str);

	try {
	    if (!matcher.matches()) {
		XCellRange range = _getTargetRange(current, x);
		XCell cell = range.getCellByPosition(0, 0);
		XCellAddressable addr = (XCellAddressable)UnoRuntime.
		    queryInterface(XCellAddressable.class, cell);
		
		XSpreadsheet sheet = _getSheet(addr.getCellAddress().Sheet);
		XNamed named = (XNamed)UnoRuntime.
		    queryInterface(XNamed.class, sheet);
		s = named.getName() + "." + x.op.str; 
	    }
	} catch (com.sun.star.lang.IndexOutOfBoundsException e) {
	    e.printStackTrace();
	}
	return s;
    }

    private void _constraintsFunction(XCellRange current,
				      CalcParser.Expression[] args) {
	try {
	    for (int i = 0; i < args.length; i++) {
		if (args[i] != null && args[i].isSymbol()) {
		    String s = _getAbsoluteRangeName(current, args[i]);
		    constraintAreaList_.add(s); // For BASIC

		    XCellRange range = 
			_getTargetRange(current, args[i]);
		    XEnumeration xenum = _cellRangeEnumeration(range);
		    while (xenum.hasMoreElements()) {
			XCell cell = (XCell)UnoRuntime.
			    queryInterface(XCell.class, xenum.nextElement());
			if (cell.getType() == CellContentType.FORMULA) {
			    String key = _getCellKey(cell);
			    CalcParser.Expression x =
				new CalcParser(cell.getFormula()).parse();
			    _constraintFunction(_getWholeRange(range), x);
			}
		    }
		}
	    }
	} catch (com.sun.star.container.NoSuchElementException e) {
	    e.printStackTrace();
	} catch (com.sun.star.lang.WrappedTargetException e) {
	    e.printStackTrace();
	}
    }

    private void _constraintFunction(XCellRange current, 
				     CalcParser.Expression x) {
	Constraint c = null;
	if (x.isOperator("=", 2)) {
	    Variable[] vs = _getTerms(current, x.args);
	    if (vs.length == 2) 
		c = new Equals(network_, vs);
	} else if (x.isOperator("<>", 2)) {
	    Variable[] vs = _getTerms(current, x.args);
	    if (vs.length == 2) 
		c = new NotEquals(network_, vs);
	    
	} else if (x.isOperator("<", 2)) {
	    Variable[] vs = _getTerms(current, x.args);
	    if (vs.length == 2) 
		c = new IntComparison
		    (network_, IntComparison.LT, vs[0], vs[1]);

	} else if (x.isOperator("<=", 2)) {
	    Variable[] vs = _getTerms(current, x.args);
	    if (vs.length == 2) 
		c = new IntComparison
		    (network_, IntComparison.LE, vs[0], vs[1]);

	} else if (x.isOperator(">", 2)) {
	    Variable[] vs = _getTerms(current, x.args);
	    if (vs.length == 2)
		c = new IntComparison
		    (network_, IntComparison.GT, vs[0], vs[1]);

	} else if (x.isOperator(">=", 2)) {
	    Variable[] vs = _getTerms(current, x.args);
	    if (vs.length == 2)
		c = new IntComparison
		    (network_, IntComparison.GE, vs[0], vs[1]);

	} else if (x.isFunctor(EQUALS) || x.isFunctor("EQUALS")) {
	    c = _equalsConstraint(current, x.args);
	} else if (x.isFunctor(NOTEQUALS) || x.isFunctor("CNOTEQUALS")) {
	    c = _notEqualsConstraint(current, x.args);
	} else if (x.isFunctor(SEQUENTIAL) || x.isFunctor("CSEQUENTIAL")) {
	    c = _sequentialConstraint(current, x.args);
	} else if (x.isFunctor(SERIALIZED) || x.isFunctor("CSERIALIZED")) {
	    c = _serializedConstraint(current, x.args);
	}
    }
    
    private Variable[] _getVariablesFromExprs(XCellRange current,
					      CalcParser.Expression[] args) {
	List<Variable> list = new ArrayList<Variable>();
	for (int i = 0; i < args.length; i++) {
	    if (args[i].isSymbol()) {
		XCellRange range = _getTargetRange(current, args[i]);
		XEnumeration xenum = _cellRangeEnumeration(range);
		
		try {
		    while (xenum.hasMoreElements()) {
			XCell cell = (XCell)UnoRuntime.
			    queryInterface(XCell.class, xenum.nextElement());
			Variable v = _getCellVariable(cell);
			if (v != null) list.add(v);
		    }
		} catch (com.sun.star.container.NoSuchElementException e) {
		    e.printStackTrace();
		} catch (com.sun.star.lang.WrappedTargetException e) {
		    e.printStackTrace();
		}
	    } else {
		int value = _cellValue(current, args[i]);
		IntVariable v = new IntVariable(network_, value);
		if (v != null) list.add(v);
	    }
	}
	Variable[] vs = new Variable[list.size()];
	return (Variable[])list.toArray(vs);
    }

    private Constraint _equalsConstraint(XCellRange current, 
					 CalcParser.Expression[] args) {
	return new Equals(network_, _getVariablesFromExprs(current, args));
    }

    private Constraint _notEqualsConstraint(XCellRange current, 
					    CalcParser.Expression[] args) {
	return new NotEquals(network_, _getVariablesFromExprs(current, args));
    }

    private Constraint _sequentialConstraint(XCellRange current,
					     CalcParser.Expression[] args) {
	int n = _countElements(current, args) / 2;
	logger_.log("sequential : " + String.valueOf(n));
	return new Sequential(network_, 
			      _getVariablesFromExprs(current, args, n + 1), 
			      _getIntValuesFromExprs(current, args, n + 1));
    }
    
    private Constraint _serializedConstraint(XCellRange current,
					     CalcParser.Expression[] args) {
	int n = _countElements(current, args) / 2;
	logger_.log("serialized : " + String.valueOf(n));
	return new Serialized(network_, 
			   _getVariablesFromExprs(current, args, n), 
			   _getIntValuesFromExprs(current, args, n));
    }

    private int _countElements(XCellRange current, 
			       CalcParser.Expression[] args) {
	int n = 0;
	for (int i = 0; i < args.length; i++) {
	    if (args[i].isSymbol()) {
		XCellRange range = _getTargetRange(current, args[i]);
		XEnumeration xenum = _cellRangeEnumeration(range);	
		
		    try {
			for (; xenum.hasMoreElements(); xenum.nextElement())
			    n++;
		    } catch (com.sun.star.container.NoSuchElementException e) {
			e.printStackTrace();
		    } catch (com.sun.star.lang.WrappedTargetException e) {
			e.printStackTrace();
		    }
	    } else { n++; }
	}
	return n;
    }

   private Variable[] _getVariablesFromExprs(XCellRange current,
					     CalcParser.Expression[] args, 
					     int num) {
       List<Variable> list = new ArrayList<Variable>();
       int n = 0;
       
       for (int i = 0; i < args.length; i++) {
	   if (args[i].isSymbol()) {
	       XCellRange range = 
		   _getTargetRange(current, args[i]);
	       XEnumeration xenum = _cellRangeEnumeration(range);
	       
	       try {
		   while (xenum.hasMoreElements()) {
		       XCell cell = (XCell)UnoRuntime.
			   queryInterface(XCell.class, xenum.nextElement());
		       Variable v = _getCellVariable(cell);
		       if (v != null && n < num) { 
			   list.add(v);
			   n++;
		       }
		   }
	       } catch (com.sun.star.container.NoSuchElementException e) {
		   e.printStackTrace();
	       } catch (com.sun.star.lang.WrappedTargetException e) {
		   e.printStackTrace();
	       }
	   } else {
	       int value = _cellValue(current, args[i]);
	       IntVariable v = new IntVariable(network_, value);
	       if (v != null && n < num) {
		   list.add(v);
		   n++;
	       }
	   }
       }
       Variable[] vs = new Variable[list.size()];
       return (Variable[])list.toArray(vs);
   }


    private int[] _getIntValuesFromExprs(XCellRange current,
					 CalcParser.Expression[] args, 
					 int num) {
	List<Integer> list = new ArrayList<Integer>();
	int n = 0;
	
	for (int i = 0; i < args.length; i++) {
	    if (args[i].isSymbol()) {
		XCellRange range = _getTargetRange(current, args[i]);
		XEnumeration xenum = _cellRangeEnumeration(range);
		
		try {
		    while (xenum.hasMoreElements()) {
			XCell cell = (XCell)UnoRuntime.
			    queryInterface(XCell.class, xenum.nextElement());
			
			Integer integer = new Integer((int)cell.getValue());
			
			if (integer != null && num <= n )
			    list.add(integer);
			n++;
		    }
		} catch (com.sun.star.container.NoSuchElementException e) {
		    e.printStackTrace();
		} catch (com.sun.star.lang.WrappedTargetException e) {
		    e.printStackTrace();
		}
	    } else {
		Integer integer = new Integer(_cellValue(current, args[i]));
		if (integer != null && num <= n)
		    list.add(integer);
		n++;
	    }
	}
	
	int[] values = new int[list.size()];
	for (int i = 0; i < list.size(); i++) {
	    values[i] = ((Integer)list.get(i)).intValue();
	}
	return values;
    }

    private void _objectiveFunction(XCellRange current, 
				    CalcParser.Expression x) {
	objective_ = _getTerm(current, x);
    }

    public void scanDocument() {
	// Scan VARIABLE
	XCellRange range = null;
	for (int i = 0; i < indexArray_.length; i++) {
	    range = _getWholeRange(_getSheet(indexArray_[i]));
	    _scanVariable(range);
	}
	// look_ahead
	for (int i = 0; i < indexArray_.length; i++) {
	    range = _getWholeRange(_getSheet(indexArray_[i]));
	    _look_ahead(range);
	}
	// Scan VARIABLES
	for (int i = 0; i < indexArray_.length; i++) {
	    range = _getWholeRange(_getSheet(indexArray_[i]));
	    _scanVariables(range);
	}
	// Scan CONSTRAINTS and OBJECTIVE etc.
	for (int i = 0; i < indexArray_.length; i++) {
	    range = _getWholeRange(_getSheet(indexArray_[i]));
	    _scanConstraints(range);
	}
    }

    // For BASIC
    private int _setDomainValue(String s) {
	int value = 0;
	s = s.toUpperCase();
	if (s.equals("MIN") || s.equals("MINIMUM")) {
	    value = IntDomain.MIN_VALUE;
	} else if (s.equals("MAX") || s.equals("MAXIMUM")) {
	    value = IntDomain.MAX_VALUE;
	} else {
	    value = Integer.parseInt(s);
	}
	return value;
    }

    private void _optionsFunction(CalcParser.Expression[] args) {
	String opt_str;
	Vector<String> solvers = new Vector<String>(); 
	    	
	for (int i = 0; i < args.length; i++) {
	    opt_str = args[i].op.str.toUpperCase();
	    
	    if (opt_str.equals("BETTER")) {
		options_ |= Solver.BETTER;
	    } else if (opt_str.equals("MAXIMIZE") || opt_str.equals("MAX")) {
		options_ |= Solver.MAXIMIZE;
	    } else if (opt_str.equals("MINIMIZE") || opt_str.equals("MIN")) {
		options_ |= Solver.MINIMIZE;
	    } else if (opt_str.equals("DEFAULT") ||
		       opt_str.equals("RW") ||
		       opt_str.equals("SA") || 
		       opt_str.equals("IBB") || 
		       opt_str.equals("TABOO")) { 
		solvers.addElement(opt_str);	
	    }
	}

	if (!solvers.isEmpty()) {
	    solvers_ = new String[solvers.size()];
	    solvers.copyInto(solvers_);
	}
    }

    private void _timeoutFunction(CalcParser.Expression[] args) {
	double timeout = Double.parseDouble(args[0].op.str);
	timeout = timeout * 1000.0;
	timeout_ = (long)timeout;
    }

    private String _getCellKey(XCell cell) {
	String key = null;
	if (cell != null) {
	    XCellAddressable addr = (XCellAddressable)UnoRuntime.
		queryInterface(XCellAddressable.class, cell);
	    CellAddress address = addr.getCellAddress();
	    key = address.Sheet + " " + address.Column + " " + address.Row;
	}
	return key;
    }

    private Variable _getCellVariable(XCell cell) {
	Variable v = null;
	String key = _getCellKey(cell);

	if (key != null && variables_.get(key) != null) {
	    v = ((CellVariable)variables_.get(key)).getVariable();
	} else {
	    v = new IntVariable(network_, (int)cell.getValue());
	}
	return v;
    }

    private static XEnumeration _cellRangeEnumeration(XCellRange range) {
	XCellRangesQuery query = (XCellRangesQuery)UnoRuntime.
	    queryInterface(XCellRangesQuery.class, range);
	
	XEnumerationAccess access = query.queryContentCells
	    (new Integer(CellFlags.VALUE + CellFlags.FORMULA).
	     shortValue()).getCells();
	
	return access.createEnumeration();
    }    

    /** 
     * An additional method for StarSuite BASIC programming.
     * @param function is a function string.
     * @return a function name string.
     */
    public String CFunctionName(String function) {
	CalcParser parser = new CalcParser(function);
	CalcParser.Expression expr = parser.parse();
	if (expr == null || expr.op == null)
	    return "";
	return expr.op.toString();
    }
    
    /**
     * An additional method for StarSuite BASIC programming.
     * @param function is a function string.
     * @return the number of the function arguments.
     */
    public int CFunctionArgc(String function) {
	CalcParser parser = new CalcParser(function);
	CalcParser.Expression expr = parser.parse();
	if (expr == null || expr.args == null) 
	    return 0;
	return expr.args.length;
    }

    /**
     * An additional method for StarSuite BASIC programming.
     * @param function is a function string.
     * @param index is the index of the argument to be returned. 
     * @return the string value of <tt>index</tt>-th argument.
     */
    public String CFunctionArg(String function, int index) {
	CalcParser parser = new CalcParser(function);
	CalcParser.Expression expr = parser.parse();
	if (expr == null || expr.args == null)
	    return "";
	return expr.args[index].toString();
    }

    /**
     * An additional method for StarSuite BASIC programming.
     * @param function is a function string.
     * @return string array of the function arguments. 
    */
    public String[] CFunctionArgs(String function) {
	CalcParser parser = new CalcParser(function);
	CalcParser.Expression expr = parser.parse();
	String[] args = new String[expr.args.length];
	for (int i = 0; i < expr.args.length; i++) 
	    args[i] = expr.args[i].toString();
	return args;
    }

    /**
     * An additional method for StarSuite BASIC programming.
     * @return the number of the ranges defined by CONSTRAINTS.
     */
    public int CAreaSize() {
	if (constraintAreaList_ == null) 
	    return 0;
	return constraintAreaList_.size();
    }
    
    /**
     * An additional method for StarSuite BASIC programming.
     * @param index is the index of the range.
     * @return <tt>index</tt>-th range name defined by CONSTRAINT.
     */
    public String CArea(int index) {
	if (constraintAreaList_ == null)
	    return "";
	return (String)constraintAreaList_.get(index);
    }

    /**
     * An additional method for StarSuite BASIC programming.
     * @param inf is the lower limit of the variable domain.
     * @param sup is the upper limit of the variable domain.
     * @return a key to identify the variable.
     */
    public String CAddVariable(String inf, String sup) {
	
	Variable v = new IntVariable
	    (network_, _setDomainValue(inf), _setDomainValue(sup));
	String key = internal_.getVariableKey();
	internal_.setVariable(key, v);
	return key;
    }

    /**
     * An additional method for StarSuite BASIC programming.
     * @param constraint is the constraint string.
     */
    public void CAddConstraint(String constraint) {
	CalcParser parser = new CalcParser(constraint);
	try {
	    CalcParser.Expression x = parser.parse();	  
	    _constraintFunction(null, x);
	} catch (java.lang.IllegalArgumentException e) {
	    e.printStackTrace();
	}
    }

    /*
     *
     */
    public void setSolution(Solution solution) {
	Pattern pattern = Pattern.compile(";[^;]*\\)\\s*$");

	for (Iterator ite = variables_.values().iterator();ite.hasNext(); ) {
	    CellVariable v = (CellVariable)ite.next();
	    int value = solution.getIntValue((IntVariable)v.getVariable());

	    if (v.getCell().getType() == CellContentType.FORMULA) {
		String formula = v.getCell().getFormula();
		
		if (formula.toUpperCase().startsWith("=" + VARIABLE + "(")) {
		    Matcher matcher = pattern.matcher(formula);
		    formula = matcher.replaceFirst(";" + value + ")");
		    v.getCell().setFormula(formula);
		}
	    } else {
		v.getCell().setValue(value);
	    }
	}
    }

    /**
     * Starts the solver.
     */
    public void start() {
	if (solver_ != null) solver_.stop();
	if (objective_ != null) network_.setObjective(objective_);
	
	if (solvers_ == null || objective_ == null) {
	    solver_ = new DefaultSolver(network_, options_);
	} else {
	    Vector<Solver> vec = new Vector<Solver>();

	    for (int i = 0; i < solvers_.length; i++) {
		if (solvers_[i].equals("DEFAULT")) {
		    vec.addElement(new DefaultSolver
				   ((Network)network_.clone(), options_));
		} else if (solvers_[i].equals("RW")) {
		    vec.addElement(new LocalSearch
				   ((Network)network_.clone(), options_));
		} else if (solvers_[i].equals("SA")) {
		    vec.addElement(new SASearch
				   ((Network)network_.clone(), options_));
		} else if (solvers_[i].equals("IBB")) {
		    vec.addElement(new IBBSearch
				   ((Network)network_.clone(), options_));
		} else if (solvers_[i].equals("TABOO")) {
		    vec.addElement(new TabooSearch
				   ((Network)network_.clone(), options_));
		}
	    }
	    
	    Solver[] solvers = new Solver[vec.size()];
	    vec.copyInto(solvers);
	    solver_ = new ParallelSolver(solvers);   
	}

	if (timeout_ < 0) 
	    solver_.start();
	else 
	    solver_.start(timeout_);
    }

    /**
     * Gets the next solution.
     */
    public void waitNext() {
	if (solver_ != null) {
	    if (timeout_ < 0) 
		solver_.waitNext();
	    else 
		solver_.waitNext(timeout_);
	}
    }

    /**
     * Finds the best solution.
     * @return solution.
     */
    public Solution findBest() {
	if (objective_ == null) return null;

	if (solver_ != null) solver_.stop();
	network_.setObjective(objective_);

	if (solvers_ == null) {
	    solver_ = new DefaultSolver(network_, options_);
	} else  {
	    Vector<Solver> vec = new Vector<Solver>();

	    for (int i = 0; i < solvers_.length; i++) {
		if (solvers_[i].equals("DEFAULT")) {
		    vec.addElement(new DefaultSolver
				   ((Network)network_.clone(), options_));
		} else if (solvers_[i].equals("RW")) {
		    vec.addElement(new LocalSearch
				   ((Network)network_.clone(), options_));
		} else if (solvers_[i].equals("SA")) {
		    vec.addElement(new SASearch
				   ((Network)network_.clone(), options_));
		} else if (solvers_[i].equals("IBB")) {
		    vec.addElement(new IBBSearch
				   ((Network)network_.clone(), options_));
		} else if (solvers_[i].equals("TABOO")) {
		    vec.addElement(new TabooSearch
				   ((Network)network_.clone(), options_));
		}
	    }
	    
	    Solver[] solvers = new Solver[vec.size()];
	    vec.copyInto(solvers);
	    solver_ = new ParallelSolver(solvers);
	    Monitor monitor = new Monitor();
	    monitor.setX(0, (int)(timeout_/1000));
	    solver_.setMonitor(monitor);
	}
   	return (timeout_ < 0)? solver_.findBest() : solver_.findBest(timeout_);
    }

    /**
     * Resumes the execution of the solver.
     */
    public void resume() {
	if (solver_ != null) solver_.resume();
    }

    /**
     * Stops the execution of the solver.
     */
    public void stop() {
	if (solver_ != null) {
	    solver_.stop();
	    solver_ = null;
	}
    }

    /**
     * Returns the last solution a solver is found.
     * @return the last solution, or null if no solution have been found.
     */
    public Solution getSolution() {
	return (solver_ != null)? solver_.getSolution() : null;
    }
} /* End of class _CreamSolver */
