/*
 * @(#)RuntimeLogger.java
 */
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.logging.FileHandler;
import java.util.logging.SimpleFormatter;

/**
 * @see CreamSolver
 * @since 0.1
 * @version 0.9, Mar 6 2005
 * @author Naoyuki Tamura
 * @author Hideaki Okamoto
 */
public class RuntimeLogger {
    private static final String LOGGER_NAME = "RuntimeLogger";
    private static final String LOG_FILE = "%h/CreamAddinsRuntime%g.log";
    private static final int LOG_LIMIT = 50000;

    private boolean active_ = false;
    private Logger logger_;
    private FileHandler handler_;
    
    /**
     * Constructs a runtime logger for cream addins system.
     * @param active makes this logger available.
     */
    public RuntimeLogger(boolean active) {
	try {
	    if (active) {
		this.active_ = active;
		this.logger_ = Logger.getLogger(LOGGER_NAME);
		this.handler_ = new FileHandler(LOG_FILE, LOG_LIMIT, 1);
		this.handler_.setFormatter(new SimpleFormatter());
		this.logger_.addHandler(handler_);
		this.logger_.setLevel(Level.ALL);
	    }
	} catch (IOException e) { e.printStackTrace(); }
    }

    public void log(String msg) {
	if (active_) logger_.log(Level.FINE, msg);
    }

    public void log(int i) {
	if (active_) logger_.log(Level.FINE, String.valueOf(i));
    }

    public void log(double d) {
	if (active_) logger_.log(Level.FINE, String.valueOf(d));
    }

    public void log(Object obj) {
	if (active_) logger_.log(Level.FINE, obj.toString());
    }
}
